@@ -47,6 +47,7 @@ Buffer::Buffer(String name, Flags flags, BufferLines lines,
 
     options().get_local_option("eolformat").set(eolformat);
     options().get_local_option("BOM").set(bom);
+    options().get_local_option("readonly").set((bool)(flags & Flags::ReadOnly));
 
     // now we may begin to record undo data
     if (not (flags & Flags::NoUndo))
@@ -69,6 +70,8 @@ void Buffer::on_registered()
         return;
     }
 
+    m_flags |= Flags::NoBufSetOption;
+
     run_hook_in_own_context(Hook::BufCreate, m_name);
 
     if (m_flags & Flags::File)
@@ -81,6 +84,7 @@ void Buffer::on_registered()
             run_hook_in_own_context(Hook::BufOpenFile, m_name);
         }
     }
+    m_flags &= ~Flags::NoBufSetOption;
 
     for (auto& option : options().flatten_options()
                       | transform(&std::unique_ptr<Option>::get)
@@ -219,40 +223,55 @@ void Buffer::reload(BufferLines lines, ByteOrderMark bom, EolFormat eolformat, F
                       [](const StringDataPtr& lhs, const StringDataPtr& rhs)
                       { return lhs->strview() == rhs->strview(); });
 
-        auto it = m_lines.begin();
+        auto read_it = m_lines.begin();
+        auto write_it = m_lines.begin();
         auto new_it = lines.begin();
-        for (auto& [op, len] : diff)
+        for (auto [op, len] : diff)
         {
+            kak_assert(read_it >= write_it);
             if (op == DiffOp::Keep)
             {
-                it += len;
+                if (read_it != write_it)
+                    std::move(read_it, read_it + len, write_it);
+                write_it += len;
+                read_it += len;
                 new_it += len;
             }
             else if (op == DiffOp::Add)
             {
-                const LineCount cur_line = (int)(it - m_lines.begin());
-
+                const LineCount cur_line = (int)(write_it - m_lines.begin());
                 for (LineCount line = 0; line < len; ++line)
                     m_current_undo_group.push_back({Modification::Insert, cur_line + line, *(new_it + (int)line)});
-
                 m_changes.push_back({Change::Insert, cur_line, cur_line + len});
-                m_lines.insert(it, new_it, new_it + len);
-                it = m_lines.begin() + (int)(cur_line + len);
+
+                if (read_it != write_it)
+                {
+                    auto count = std::min(len, static_cast<int>(read_it - write_it));
+                    write_it = std::copy(new_it, new_it + count, write_it);
+                    new_it += count;
+                    if (len == count)
+                        continue;
+                    len -= count;
+                }
+
+                auto read_pos = read_it - m_lines.begin();
+                write_it = m_lines.insert(write_it, new_it, new_it + len) + len;
+                read_it = m_lines.begin() + read_pos + len;
                 new_it += len;
             }
             else if (op == DiffOp::Remove)
             {
-                const LineCount cur_line = (int)(it - m_lines.begin());
-
+                const LineCount cur_line = (int)(write_it - m_lines.begin());
                 for (LineCount line = len-1; line >= 0; --line)
                     m_current_undo_group.push_back({
                         Modification::Erase, cur_line + line,
-                        m_lines.get_storage(cur_line + line)});
+                        *(read_it + (size_t)line)});
 
-                it = m_lines.erase(it, it + len);
+                read_it += len;
                 m_changes.push_back({ Change::Erase, cur_line, cur_line + len });
             }
         }
+        m_lines.erase(write_it, m_lines.end());
     }
 
     commit_undo_group();
@@ -305,11 +324,10 @@ bool Buffer::redo(size_t count)
 {
     throw_if_read_only();
 
-    if (current_history_node().redo_child == HistoryId::Invalid)
+    if (current_history_node().redo_child == HistoryId::Invalid or
+        not m_current_undo_group.empty())
         return false;
 
-    kak_assert(m_current_undo_group.empty());
-
     while (count-- != 0 and current_history_node().redo_child != HistoryId::Invalid)
     {
         m_history_id = current_history_node().redo_child;
@@ -415,14 +433,14 @@ BufferRange Buffer::do_insert(BufferCoord pos, StringView content)
         if (content[i] == '\n')
         {
             StringView line = content.substr(start, i + 1 - start);
-            new_lines.push_back(start == 0 ? StringData::create({prefix, line}) : StringData::create({line}));
+            new_lines.push_back(start == 0 ? StringData::create(prefix, line) : StringData::create(line));
             start = i + 1;
         }
     }
     if (start == 0)
-        new_lines.push_back(StringData::create({prefix, content, suffix}));
+        new_lines.push_back(StringData::create(prefix, content, suffix));
     else if (start != content.length() or not suffix.empty())
-        new_lines.push_back(StringData::create({content.substr(start), suffix}));
+        new_lines.push_back(StringData::create(content.substr(start), suffix));
 
     auto line_it = m_lines.begin() + (int)pos.line;
     auto new_lines_it = new_lines.begin();
@@ -451,7 +469,7 @@ BufferCoord Buffer::do_erase(BufferCoord begin, BufferCoord end)
     StringView prefix = m_lines[begin.line].substr(0, begin.column);
     StringView suffix = end.line == line_count() ? StringView{} : m_lines[end.line].substr(end.column);
 
-    auto new_line = (not prefix.empty() or not suffix.empty()) ? StringData::create({prefix, suffix}) : StringDataPtr{};
+    auto new_line = (not prefix.empty() or not suffix.empty()) ? StringData::create(prefix, suffix) : StringDataPtr{};
     m_lines.erase(m_lines.begin() + (int)begin.line, m_lines.begin() + (int)end.line);
 
     m_changes.push_back({ Change::Erase, begin, end });
@@ -555,17 +573,17 @@ void Buffer::notify_saved(FsStatus status)
     m_fs_status = status;
 }
 
-BufferCoord Buffer::advance(BufferCoord coord, ByteCount count) const
+BufferCoord Buffer::advance(ArrayView<const StringDataPtr> lines, BufferCoord coord, ByteCount count)
 {
     if (count > 0)
     {
         auto line = coord.line;
         count += coord.column;
-        while (count >= m_lines[line].length())
+        while (count >= lines[(size_t)line]->length)
         {
-            count -= m_lines[line++].length();
-            if (line == line_count())
-                return end_coord();
+            count -= lines[(size_t)line++]->length;
+            if ((size_t)line == lines.size())
+                return line;
         }
         return { line, count };
     }
@@ -577,7 +595,7 @@ BufferCoord Buffer::advance(BufferCoord coord, ByteCount count) const
         {
             if (--line < 0)
                 return {0, 0};
-            count += m_lines[line].length();
+            count += lines[(size_t)line]->length;
         }
         return { line, count };
     }
@@ -622,6 +640,9 @@ const FsStatus& Buffer::fs_status() const
 
 void Buffer::on_option_changed(const Option& option)
 {
+    if (m_flags & Flags::NoBufSetOption)
+        return;
+
     if (option.name() == "readonly")
     {
         if (option.get<bool>())
@@ -676,7 +697,7 @@ String Buffer::debug_description() const
 
 UnitTest test_buffer{[]()
 {
-    auto make_lines = [](auto&&... lines) { return BufferLines{StringData::create({lines})...}; };
+    auto make_lines = [](auto&&... lines) { return BufferLines{StringData::create(lines)...}; };
 
     Buffer empty_buffer("empty", Buffer::Flags::None, make_lines("\n"));
 
@@ -723,7 +744,7 @@ UnitTest test_buffer{[]()
 
 UnitTest test_undo{[]()
 {
-    auto make_lines = [](auto&&... lines) { return BufferLines{StringData::create({lines})...}; };
+    auto make_lines = [](auto&&... lines) { return BufferLines{StringData::create(lines)...}; };
 
     Buffer buffer("test", Buffer::Flags::None, make_lines("allo ?\n", "mais que fais la police\n", " hein ?\n", " youpi\n"));
     auto pos = buffer.end_coord();
