@@ -1,6 +1,7 @@
 #include "buffer_utils.hh"
 
 #include "buffer_manager.hh"
+#include "coord.hh"
 #include "event_manager.hh"
 #include "file.hh"
 #include "selection.hh"
@@ -97,12 +98,12 @@ static BufferLines parse_lines(const char* pos, const char* end, EolFormat eolfo
         if ((eol - pos) >= std::numeric_limits<int>::max())
             throw runtime_error("line is too long");
 
-        lines.emplace_back(StringData::create({{pos, eol - (eolformat == EolFormat::Crlf and eol != end ? 1 : 0)}, "\n"}));
+        lines.emplace_back(StringData::create(StringView{pos, eol - (eolformat == EolFormat::Crlf and eol != end ? 1 : 0)}, "\n"));
         pos = eol + 1;
     }
 
     if (lines.empty())
-        lines.emplace_back(StringData::create({"\n"}));
+        lines.emplace_back(StringData::create("\n"));
 
     return lines;
 }
@@ -132,15 +133,11 @@ decltype(auto) parse_file(StringView filename, Func&& func)
     }
 
     bool has_crlf = false, has_lf = false;
-    for (auto it = pos; it != end; ++it)
-    {
-        if (*it == '\n')
-            ((it != pos and *(it-1) == '\r') ? has_crlf : has_lf) = true;
-    }
-    const bool crlf = has_crlf and not has_lf;
-    auto eolformat = crlf ? EolFormat::Crlf : EolFormat::Lf;
+    for (auto it = std::find(pos, end, '\n'); it != end; it = std::find(it+1, end, '\n'))
+        ((it != pos and *(it-1) == '\r') ? has_crlf : has_lf) = true;
+    auto eolformat = (has_crlf and not has_lf) ? EolFormat::Crlf : EolFormat::Lf;
 
-    FsStatus fs_status{file.st.st_mtim, file.st.st_size, hash_data(file.data, file.st.st_size)};
+    FsStatus fs_status{file.st.st_mtim, file.st.st_size, murmur3(file.data, file.st.st_size)};
     return func(parse_lines(pos, end, eolformat), bom, eolformat, fs_status);
 }
 
@@ -169,7 +166,97 @@ void reload_file_buffer(Buffer& buffer)
     buffer.flags() &= ~Buffer::Flags::New;
 }
 
-Buffer* create_fifo_buffer(String name, int fd, Buffer::Flags flags, bool scroll)
+void write_buffer_to_fd(Buffer& buffer, int fd)
+{
+    auto eolformat = buffer.options()["eolformat"].get<EolFormat>();
+    StringView eoldata;
+    if (eolformat == EolFormat::Crlf)
+        eoldata = "\r\n";
+    else
+        eoldata = "\n";
+
+
+    BufferedWriter<false> writer{fd};
+    if (buffer.options()["BOM"].get<ByteOrderMark>() == ByteOrderMark::Utf8)
+        writer.write("\xEF\xBB\xBF");
+
+    for (LineCount i = 0; i < buffer.line_count(); ++i)
+    {
+        // end of lines are written according to eolformat but always
+        // stored as \n
+        StringView linedata = buffer[i];
+        writer.write(linedata.substr(0, linedata.length()-1));
+        writer.write(eoldata);
+    }
+}
+
+void write_buffer_to_file(Buffer& buffer, StringView filename,
+                          WriteMethod method, WriteFlags flags)
+{
+    auto zfilename = filename.zstr();
+    struct stat st;
+
+    bool replace = method == WriteMethod::Replace;
+    bool force = flags & WriteFlags::Force;
+
+    if ((replace or force) and (::stat(zfilename, &st) != 0 or
+                                (st.st_mode & S_IFMT) != S_IFREG))
+    {
+        force = false;
+        replace = false;
+    }
+
+    if (force and ::chmod(zfilename, st.st_mode | S_IWUSR) < 0)
+        throw runtime_error(format("unable to change file permissions: {}", strerror(errno)));
+
+    char temp_filename[PATH_MAX];
+    const int fd = replace ? open_temp_file(filename, temp_filename)
+                           : create_file(zfilename);
+    if (fd == -1)
+    {
+        auto saved_errno = errno;
+        if (force)
+            ::chmod(zfilename, st.st_mode);
+        throw file_access_error(filename, strerror(saved_errno));
+    }
+
+    {
+        auto close_fd = on_scope_end([fd]{ close(fd); });
+        write_buffer_to_fd(buffer, fd);
+        if (flags & WriteFlags::Sync)
+            ::fsync(fd);
+    }
+
+    if (replace and geteuid() == 0 and ::chown(temp_filename, st.st_uid, st.st_gid) < 0)
+        throw runtime_error(format("unable to set replacement file ownership: {}", strerror(errno)));
+    if (replace and ::chmod(temp_filename, st.st_mode) < 0)
+        throw runtime_error(format("unable to set replacement file permissions: {}", strerror(errno)));
+    if (force and not replace and ::chmod(zfilename, st.st_mode) < 0)
+        throw runtime_error(format("unable to restore file permissions: {}", strerror(errno)));
+
+    if (replace and rename(temp_filename, zfilename) != 0)
+    {
+        if (force)
+            ::chmod(zfilename, st.st_mode);
+        throw runtime_error("replacing file failed");
+    }
+
+    if ((buffer.flags() & Buffer::Flags::File) and
+        real_path(filename) == real_path(buffer.name()))
+        buffer.notify_saved(get_fs_status(real_path(filename)));
+}
+
+void write_buffer_to_backup_file(Buffer& buffer)
+{
+    const int fd = open_temp_file(buffer.name());
+    if (fd >= 0)
+    {
+        write_buffer_to_fd(buffer, fd);
+        close(fd);
+    }
+}
+
+Buffer* create_fifo_buffer(String name, int fd, Buffer::Flags flags, AutoScroll scroll)
 {
     static ValueId fifo_watcher_id = get_free_value_id();
 
@@ -178,16 +265,18 @@ Buffer* create_fifo_buffer(String name, int fd, Buffer::Flags flags, bool scroll
     if (buffer)
     {
         buffer->flags() |= Buffer::Flags::NoUndo | flags;
-        buffer->reload({StringData::create({"\n"})}, ByteOrderMark::None, EolFormat::Lf, {InvalidTime, {}, {}});
+        buffer->values().clear();
+        buffer->reload({StringData::create("\n")}, ByteOrderMark::None, EolFormat::Lf, {InvalidTime, {}, {}});
+        buffer_manager.make_latest(*buffer);
     }
     else
         buffer = buffer_manager.create_buffer(
             std::move(name), flags | Buffer::Flags::Fifo | Buffer::Flags::NoUndo,
-            {StringData::create({"\n"})}, ByteOrderMark::None, EolFormat::Lf, {InvalidTime, {}, {}});
+            {StringData::create("\n")}, ByteOrderMark::None, EolFormat::Lf, {InvalidTime, {}, {}});
 
     struct FifoWatcher : FDWatcher
     {
-        FifoWatcher(int fd, Buffer& buffer, bool scroll)
+        FifoWatcher(int fd, Buffer& buffer, AutoScroll scroll)
             : FDWatcher(fd, FdEvents::Read, EventMode::Normal,
                         [](FDWatcher& watcher, FdEvents, EventMode mode) {
                             if (mode == EventMode::Normal)
@@ -204,7 +293,7 @@ Buffer* create_fifo_buffer(String name, int fd, Buffer::Flags flags, bool scroll
             m_buffer.flags() &= ~(Buffer::Flags::Fifo | Buffer::Flags::NoUndo);
         }
 
-        void read_fifo() const
+        void read_fifo()
         {
             kak_assert(m_buffer.flags() & Buffer::Flags::Fifo);
 
@@ -216,7 +305,7 @@ Buffer* create_fifo_buffer(String name, int fd, Buffer::Flags flags, bool scroll
             bool closed = false;
             size_t loop = 0;
             char data[buffer_size];
-            BufferCoord insert_coord = m_buffer.back_coord();
+            Optional<BufferCoord> insert_begin;
             const int fifo = fd();
 
             {
@@ -233,35 +322,51 @@ Buffer* create_fifo_buffer(String name, int fd, Buffer::Flags flags, bool scroll
                     }
 
                     auto pos = m_buffer.back_coord();
-                    const bool prevent_scrolling = pos == BufferCoord{0,0} and not m_scroll;
-                    if (prevent_scrolling)
+                    const bool is_first = pos == BufferCoord{0,0};
+                    if ((m_scroll == AutoScroll::No and (is_first or m_had_trailing_newline))
+                        or (m_scroll == AutoScroll::NotInitially and is_first))
                         pos = m_buffer.next(pos);
 
-                    m_buffer.insert(pos, StringView(data, data+count));
+                    auto inserted_range = m_buffer.insert(pos, StringView(data, data+count));
+                    if (not insert_begin)
+                        insert_begin = inserted_range.begin;
+                    pos = inserted_range.end;
 
-                    if (prevent_scrolling)
+                    bool have_trailing_newline = (data[count-1] == '\n');
+                    if (m_scroll != AutoScroll::Yes)
                     {
-                        m_buffer.erase({0,0}, m_buffer.next({0,0}));
-                        // in the other case, the buffer will have automatically
-                        // inserted a \n to guarantee its invariant.
-                        if (data[count-1] == '\n')
-                            m_buffer.insert(m_buffer.end_coord(), "\n");
+                        if (is_first)
+                        {
+                            m_buffer.erase({0,0}, m_buffer.next({0,0}));
+                            --insert_begin->line;
+                            if (m_scroll == AutoScroll::NotInitially and have_trailing_newline)
+                                m_buffer.insert(m_buffer.end_coord(), "\n");
+                        }
+                        else if (m_scroll == AutoScroll::No and
+                                 not m_had_trailing_newline and have_trailing_newline)
+                            m_buffer.erase(m_buffer.prev(pos), pos);
                     }
+                    m_had_trailing_newline = have_trailing_newline;
                 }
                 while (++loop < max_loop  and fd_readable(fifo));
             }
 
-            if (insert_coord != m_buffer.back_coord())
+            if (insert_begin)
+            {
+                auto insert_back = (m_had_trailing_newline and m_scroll == AutoScroll::No)
+                                 ? m_buffer.back_coord() : m_buffer.prev(m_buffer.back_coord());
                 m_buffer.run_hook_in_own_context(
                     Hook::BufReadFifo,
-                    selection_to_string(ColumnType::Byte, m_buffer, {insert_coord, m_buffer.back_coord()}));
+                    selection_to_string(ColumnType::Byte, m_buffer, {*insert_begin, insert_back}));
+            }
 
             if (closed)
                 m_buffer.values().erase(fifo_watcher_id); // will delete this
         }
 
         Buffer& m_buffer;
-        bool m_scroll;
+        AutoScroll m_scroll;
+        bool m_had_trailing_newline = false;
     };
 
     buffer->values()[fifo_watcher_id] = Value(Meta::Type<FifoWatcher>{}, fd, *buffer, scroll);
@@ -271,36 +376,6 @@ Buffer* create_fifo_buffer(String name, int fd, Buffer::Flags flags, bool scroll
     return buffer;
 }
 
-void write_to_debug_buffer(StringView str)
-{
-    if (not BufferManager::has_instance())
-    {
-        write(2, str);
-        write(2, "\n");
-        return;
-    }
-
-    constexpr StringView debug_buffer_name = "*debug*";
-    // Try to ensure we keep an empty line at the end of the debug buffer
-    // where the user can put its cursor to scroll with new messages
-    const bool eol_back = not str.empty() and str.back() == '\n';
-    if (Buffer* buffer = BufferManager::instance().get_buffer_ifp(debug_buffer_name))
-    {
-        buffer->flags() &= ~Buffer::Flags::ReadOnly;
-        auto restore = on_scope_end([buffer] { buffer->flags() |= Buffer::Flags::ReadOnly; });
-
-        buffer->insert(buffer->back_coord(), eol_back ? str : str + "\n");
-    }
-    else
-    {
-        String line = str + (eol_back ? "\n" : "\n\n");
-        create_buffer_from_string(
-            debug_buffer_name.str(), Buffer::Flags::NoUndo | Buffer::Flags::Debug | Buffer::Flags::ReadOnly,
-            line);
-    }
-}
-
-
 auto to_string(Buffer::HistoryId id)
 {
     using Result = decltype(to_string(size_t{}));
