@@ -1,16 +1,17 @@
 #include "client.hh"
 
-#include "face_registry.hh"
+#include "clock.hh"
 #include "context.hh"
-#include "buffer_manager.hh"
 #include "buffer_utils.hh"
+#include "debug.hh"
 #include "file.hh"
 #include "remote.hh"
 #include "option.hh"
 #include "option_types.hh"
 #include "client_manager.hh"
-#include "command_manager.hh"
 #include "event_manager.hh"
+#include "shell_manager.hh"
+#include "command_manager.hh"
 #include "user_interface.hh"
 #include "window.hh"
 #include "hash_map.hh"
@@ -62,7 +63,7 @@ Client::Client(std::unique_ptr<UserInterface>&& ui,
         else if (key.modifiers & Key::Modifiers::Resize)
         {
             m_window->set_dimensions(key.coord());
-            force_redraw();
+            force_redraw(true);
         }
         else
             m_pending_keys.push_back(key);
@@ -109,7 +110,10 @@ bool Client::process_pending_inputs()
             else if (key == Key::FocusOut)
                 context().hooks().run_hook(Hook::FocusOut, context().name(), context());
             else
-                m_input_handler.handle_key(key);
+            {
+                context().ensure_cursor_visible = true;
+                m_input_handler.handle_key(key, /*synthesized=*/false);
+            }
 
             context().hooks().run_hook(Hook::RawKey, to_string(key), context());
         }
@@ -127,6 +131,7 @@ void Client::print_status(DisplayLine status_line)
 {
     m_status_line = std::move(status_line);
     m_ui_pending |= StatusLine;
+    m_pending_clear &= ~PendingClear::StatusLine;
 }
 
 
@@ -162,11 +167,16 @@ DisplayLine Client::generate_mode_line() const
     DisplayLine modeline;
     try
     {
+        auto [mode_info_line, normal_params] = context().client().input_handler().mode_info();
         const String& modelinefmt = context().options()["modelinefmt"].get<String>();
-        HashMap<String, DisplayLine> atoms{{ "mode_info", context().client().input_handler().mode_line() },
+        HashMap<String, DisplayLine> atoms{{ "mode_info",  mode_info_line},
                                            { "context_info", {generate_context_info(context()),
                                                               context().faces()["Information"]}}};
-        auto expanded = expand(modelinefmt, context(), ShellContext{},
+        ShellContext shell_context{{}, {
+            {"register", normal_params ? StringView{normal_params->reg}.str() : ""},
+            {"count", normal_params ? String{to_string(normal_params->count)} : ""},
+        }};
+        auto expanded = expand(modelinefmt, context(), shell_context,
                                [](String s) { return escape(s, '{', '\\'); });
         modeline = parse_display_line(expanded, context().faces(), atoms);
     }
@@ -184,6 +194,9 @@ void Client::change_buffer(Buffer& buffer, Optional<FunctionRef<void()>> set_sel
     if (m_buffer_reload_dialog_opened)
         close_buffer_reload_dialog();
 
+    buffer.flags() |= Buffer::Flags::Locked;
+    OnScopeEnd unlock{[&] { buffer.flags() &= ~Buffer::Flags::Locked; }};
+
     auto& client_manager = ClientManager::instance();
     WindowAndSelections ws = client_manager.get_free_window(buffer);
 
@@ -210,7 +223,7 @@ void Client::change_buffer(Buffer& buffer, Optional<FunctionRef<void()>> set_sel
     m_ui->set_ui_options(m_window->options()["ui_options"].get<UserInterface::Options>());
 
     m_window->hooks().run_hook(Hook::WinDisplay, buffer.name(), context());
-    force_redraw();
+    force_redraw(true);
 }
 
 static bool is_inline(InfoStyle style)
@@ -247,7 +260,7 @@ void Client::redraw_ifn()
     if ((m_ui_pending & MenuShow) or update_menu_anchor)
     {
         auto anchor = m_menu.style == MenuStyle::Inline ?
-            window.display_position(m_menu.anchor) : DisplayCoord{};
+            window.display_coord(m_menu.anchor) : DisplayCoord{};
         if (not (m_ui_pending & MenuShow) and m_menu.ui_anchor != anchor)
             m_ui_pending |= anchor ? (MenuShow | MenuSelect) : MenuHide;
         m_menu.ui_anchor = anchor;
@@ -267,7 +280,7 @@ void Client::redraw_ifn()
     if ((m_ui_pending & InfoShow) or update_info_anchor)
     {
         auto anchor = is_inline(m_info.style) ?
-             window.display_position(m_info.anchor) : DisplayCoord{};
+             window.display_coord(m_info.anchor) : DisplayCoord{};
         if (not (m_ui_pending & MenuShow) and m_info.ui_anchor != anchor)
             m_ui_pending |= anchor ? InfoShow : InfoHide;
         m_info.ui_anchor = anchor;
@@ -275,7 +288,8 @@ void Client::redraw_ifn()
 
     if (m_ui_pending & InfoShow and m_info.ui_anchor)
         m_ui->info_show(m_info.title, m_info.content, *m_info.ui_anchor,
-                        faces["Information"], m_info.style);
+                        faces[(is_inline(m_info.style) || m_info.style == InfoStyle::MenuDoc)
+                        ? "InlineInformation" : "Information"], m_info.style);
     if (m_ui_pending & InfoHide)
         m_ui->info_hide();
 
@@ -289,11 +303,14 @@ void Client::redraw_ifn()
     m_ui_pending = 0;
 }
 
-void Client::force_redraw()
+void Client::force_redraw(bool full)
 {
-    m_ui_pending |= Refresh | Draw | StatusLine |
-        (m_menu.items.empty() ? MenuHide : MenuShow | MenuSelect) |
-        (m_info.content.empty() ? InfoHide : InfoShow);
+    if (full)
+        m_ui_pending |= Refresh | Draw | StatusLine |
+            (m_menu.items.empty() ? MenuHide : MenuShow | MenuSelect) |
+            (m_info.content.empty() ? InfoHide : InfoShow);
+    else
+        m_ui_pending |= Draw;
 }
 
 void Client::reload_buffer()
@@ -387,7 +404,7 @@ void Client::check_if_buffer_needs_reloading()
             return;
 
         if (MappedFile fd{filename};
-            fd.st.st_size == status.file_size and hash_data(fd.data, fd.st.st_size) == status.hash)
+            fd.st.st_size == status.file_size and murmur3(fd.data, fd.st.st_size) == status.hash)
             return;
 
         if (reload == Autoreload::Ask)
@@ -422,10 +439,8 @@ StringView Client::get_env_var(StringView name) const
 void Client::on_option_changed(const Option& option)
 {
     if (option.name() == "ui_options")
-    {
         m_ui->set_ui_options(option.get<UserInterface::Options>());
-        m_ui_pending |= Draw;
-    }
+    m_ui_pending |= Draw; // a highlighter might depend on the option, so we need to redraw
 }
 
 void Client::menu_show(Vector<DisplayLine> choices, BufferCoord anchor, MenuStyle style)
@@ -457,6 +472,7 @@ void Client::info_show(DisplayLine title, DisplayLineList content, BufferCoord a
     m_info = Info{ std::move(title), std::move(content), anchor, {}, style };
     m_ui_pending |= InfoShow;
     m_ui_pending &= ~InfoHide;
+    m_pending_clear &= ~PendingClear::Info;
 }
 
 void Client::info_show(StringView title, StringView content, BufferCoord anchor, InfoStyle style)
@@ -480,4 +496,52 @@ void Client::info_hide(bool even_modal)
     m_ui_pending &= ~InfoShow;
 }
 
+void Client::schedule_clear()
+{
+    if (not (m_ui_pending & InfoShow))
+        m_pending_clear |= PendingClear::Info;
+    if (not (m_ui_pending & StatusLine))
+        m_pending_clear |= PendingClear::StatusLine;
+}
+
+void Client::clear_pending()
+{
+    if (m_pending_clear & PendingClear::StatusLine)
+        print_status({});
+    if (m_pending_clear & PendingClear::Info)
+        info_hide();
+    m_pending_clear = PendingClear::None;
+}
+
+constexpr std::chrono::seconds wait_timeout{1};
+
+BusyIndicator::BusyIndicator(const Context& context,
+                             std::function<DisplayLine(std::chrono::seconds)> status_message,
+                             TimePoint wait_time)
+    : m_context(context),
+      m_timer{wait_time + wait_timeout,
+        [this, status_message = std::move(status_message), wait_time](Timer& timer) {
+            if (not m_context.has_client())
+                return;
+            using namespace std::chrono;
+            const auto now = Clock::now();
+            timer.set_next_date(now + wait_timeout);
+
+            auto& client = m_context.client();
+            if (not m_previous_status)
+                m_previous_status = client.current_status();
+
+            client.print_status(status_message(duration_cast<seconds>(now - wait_time)));
+            client.redraw_ifn();
+        }, EventMode::Urgent} {}
+
+BusyIndicator::~BusyIndicator()
+{
+    if (m_previous_status and std::uncaught_exceptions() == 0) // restore the status line
+    {
+        m_context.print_status(std::move(*m_previous_status));
+        m_context.client().redraw_ifn();
+    }
+}
+
 }
