@@ -1,7 +1,7 @@
 #include "remote.hh"
 
-#include "buffer_manager.hh"
 #include "buffer_utils.hh"
+#include "debug.hh"
 #include "client_manager.hh"
 #include "command_manager.hh"
 #include "display_buffer.hh"
@@ -619,21 +619,21 @@ const String& session_directory()
     return session_dir;
 }
 
-String session_path(StringView session)
+String session_path(StringView session, bool assume_valid)
 {
-    if (not all_of(session, is_identifier))
+    if (not assume_valid and not all_of(session, is_identifier))
         throw runtime_error{format("invalid session name: '{}'", session)};
-    return format("{}/{}", session_directory(), session);
+    String path = format("{}/{}", session_directory(), session);
+    if (not assume_valid and path.length() + 1 > sizeof sockaddr_un{}.sun_path)
+        throw runtime_error{format("socket path too long: '{}'", path)};
+    return path;
 }
 
 static sockaddr_un session_addr(StringView session)
 {
     sockaddr_un addr;
     addr.sun_family = AF_UNIX;
-    String path = session_path(session);
-    if (path.length() + 1 > sizeof addr.sun_path)
-        throw runtime_error{format("socket path too long: '{}'", path)};
-    strcpy(addr.sun_path, path.c_str());
+    strcpy(addr.sun_path, session_path(session).c_str());
     return addr;
 }
 
@@ -806,8 +806,8 @@ private:
                 auto env_vars = m_reader.read<HashMap<String, String, MemoryDomain::EnvVars>>();
 
                 if (auto stdin_fd = m_reader.ancillary_fd())
-                    create_fifo_buffer(generate_buffer_name("*stdin-{}*"), *stdin_fd, Buffer::Flags::None);
-
+                    create_fifo_buffer(generate_buffer_name("*stdin-{}*"), *stdin_fd, Buffer::Flags::None,
+                                       AutoScroll::NotInitially);
                 auto* ui = new RemoteUI{sock, dimensions};
                 ClientManager::instance().create_client(
                     std::unique_ptr<UserInterface>(ui), pid, std::move(name),
@@ -889,7 +889,7 @@ Server::Server(String session_name, bool is_daemon)
 
 bool Server::rename_session(StringView name)
 {
-    String old_socket_file = session_path(m_session);
+    String old_socket_file = session_path(m_session, true);
     String new_socket_file = session_path(name);
 
     if (file_exists(new_socket_file))
@@ -906,7 +906,7 @@ void Server::close_session(bool do_unlink)
 {
     if (do_unlink)
     {
-        String socket_file = session_path(m_session);
+        String socket_file = session_path(m_session, true);
         unlink(socket_file.c_str());
     }
     m_listener->close_fd();
