@@ -1,14 +1,14 @@
 #include "terminal_ui.hh"
 
-#include "buffer_utils.hh"
 #include "display_buffer.hh"
 #include "event_manager.hh"
 #include "exception.hh"
 #include "file.hh"
 #include "keys.hh"
 #include "ranges.hh"
-#include "string_utils.hh"
+#include "format.hh"
 #include "diff.hh"
+#include "string_utils.hh"
 
 #include <algorithm>
 
@@ -30,8 +30,8 @@ static String fix_atom_text(StringView str)
     auto pos = str.begin();
     for (auto it = str.begin(), end = str.end(); it != end; ++it)
     {
-        char c = *it;
-        if (c >= 0 and c <= 0x1F)
+        unsigned char c = *it;
+        if (c <= 0x1F)
         {
             res += StringView{pos, it};
             res += String{Codepoint{(uint32_t)(0x2400 + c)}};
@@ -68,8 +68,7 @@ struct TerminalUI::Window::Line
             text.resize(it - text.begin(), 0);
         }
 
-        friend bool operator==(const Atom& lhs, const Atom& rhs) { return lhs.text == rhs.text and lhs.skip == rhs.skip and lhs.face == rhs.face; }
-        friend bool operator!=(const Atom& lhs, const Atom& rhs) { return not (lhs == rhs); }
+        friend bool operator==(const Atom& lhs, const Atom& rhs) = default;
         friend size_t hash_value(const Atom& atom) { return hash_values(atom.text, atom.skip, atom.face); }
     };
 
@@ -225,7 +224,7 @@ void TerminalUI::Screen::set_face(const Face& face, Writer& writer)
     static constexpr int fg_table[]{ 39, 30, 31, 32, 33, 34, 35, 36, 37, 90, 91, 92, 93, 94, 95, 96, 97 };
     static constexpr int bg_table[]{ 49, 40, 41, 42, 43, 44, 45, 46, 47, 100, 101, 102, 103, 104, 105, 106, 107 };
     static constexpr int ul_table[]{ 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 };
-    static constexpr const char* attr_table[]{ "0", "4", "4:3", "7", "5", "1", "2", "3", "9" };
+    static constexpr const char* attr_table[]{ "0", "4", "4:3", "21", "7", "5", "1", "2", "3", "9" };
 
     auto set_color = [&](bool fg, const Color& color, bool join) {
         if (join)
@@ -378,7 +377,7 @@ void TerminalUI::Screen::output(bool force, bool synchronized, Writer& writer)
     {
         for (int line = 0; line < (int)size.line; ++line)
         {
-            auto hash = hash_line(lines[line]); 
+            auto hash = hash_line(lines[line]);
             if (hash == hashes[line])
                 continue;
             hashes[line] = hash;
@@ -429,7 +428,7 @@ static constexpr StringView assistant_dilbert[] =
 template<typename T> T sq(T x) { return x * x; }
 
 static sig_atomic_t resize_pending = 0;
-static sig_atomic_t stdin_closed = 0;
+static sig_atomic_t terminal_hungup = 0;
 
 template<sig_atomic_t* signal_flag>
 static void signal_handler(int)
@@ -465,7 +464,7 @@ TerminalUI::TerminalUI()
     enable_mouse(true);
 
     set_signal_handler(SIGWINCH, &signal_handler<&resize_pending>);
-    set_signal_handler(SIGHUP, &signal_handler<&stdin_closed>);
+    set_signal_handler(SIGHUP, &signal_handler<&terminal_hungup>);
     set_signal_handler(SIGTSTP, [](int){ TerminalUI::instance().suspend(); });
 
     check_resize(true);
@@ -474,11 +473,13 @@ TerminalUI::TerminalUI()
 
 TerminalUI::~TerminalUI()
 {
-    enable_mouse(false);
-    restore_terminal();
-    tcsetattr(STDIN_FILENO, TCSAFLUSH, &m_original_termios);
+    if (not terminal_hungup) {
+        enable_mouse(false);
+        restore_terminal();
+        tcsetattr(STDIN_FILENO, TCSAFLUSH, &m_original_termios);
+    }
     set_signal_handler(SIGWINCH, SIG_DFL);
-    set_signal_handler(SIGHUP, SIG_DFL);
+    set_signal_handler(SIGHUP, SIG_IGN);
     set_signal_handler(SIGTSTP, SIG_DFL);
 }
 
@@ -608,16 +609,19 @@ void TerminalUI::draw_status(const DisplayLine& status_line,
     if (m_set_title)
     {
         Writer writer{STDOUT_FILENO};
-        constexpr char suffix[] = " - Kakoune\007";
         writer.write("\033]2;");
-        // Fill title escape sequence buffer, removing non ascii characters
-        for (auto& atom : mode_line)
-        {
-            const auto str = atom.content();
+        auto write_escaped = [&](StringView str) {
             for (auto it = str.begin(), end = str.end(); it != end; utf8::to_next(it, end))
                 writer.write((*it >= 0x20 and *it <= 0x7e) ? *it : '?');
+        };
+        if (not m_title)
+        {
+            for (auto& atom : mode_line)
+                write_escaped(atom.content());
         }
-        writer.write(suffix);
+        else
+            write_escaped(*m_title);
+        writer.write(" - Kakoune\007");
     }
 
     m_dirty = true;
@@ -666,10 +670,10 @@ void TerminalUI::check_resize(bool force)
 
 Optional<Key> TerminalUI::get_next_key()
 {
-    if (stdin_closed)
+    if (terminal_hungup)
     {
         set_signal_handler(SIGWINCH, SIG_DFL);
-        set_signal_handler(SIGHUP, SIG_DFL);
+        set_signal_handler(SIGHUP, SIG_IGN);
         if (m_window)
             m_window.destroy();
         m_stdin_watcher.disable();
@@ -691,7 +695,7 @@ Optional<Key> TerminalUI::get_next_key()
         if (unsigned char c = 0; read(STDIN_FILENO, &c, 1) == 1)
             return c;
 
-        stdin_closed = 1;
+        terminal_hungup = 1;
         return {};
     };
 
@@ -701,8 +705,8 @@ Optional<Key> TerminalUI::get_next_key()
 
     static constexpr auto control = [](char c) { return c & 037; };
 
-    static auto convert = [this](Codepoint c) -> Codepoint {
-        if (c == control('m') or c == control('j'))
+    auto convert = [this](Codepoint c) -> Codepoint {
+        if (c == control('m'))
             return Key::Return;
         if (c == control('i'))
             return Key::Tab;
@@ -716,7 +720,7 @@ Optional<Key> TerminalUI::get_next_key()
             return Key::Escape;
         return c;
     };
-    static auto parse_key = [](unsigned char c) -> Key {
+    auto parse_key = [&convert](unsigned char c) -> Key {
         if (Codepoint cp = convert(c); cp > 255)
             return Key{cp};
         // Special case: you can type NUL with Ctrl-2 or Ctrl-Shift-2 or
@@ -755,7 +759,7 @@ Optional<Key> TerminalUI::get_next_key()
         return mod;
     };
 
-    auto parse_csi = [this]() -> Optional<Key> {
+    auto parse_csi = [this, &convert]() -> Optional<Key> {
         auto next_char = [] { return get_char().value_or((unsigned char)0xff); };
         int params[16][4] = {};
         auto c = next_char();
@@ -797,9 +801,8 @@ Optional<Key> TerminalUI::get_next_key()
             return Key{mod | Key::to_modifier(button), coord};
         };
 
-        auto mouse_scroll = [this](Key::Modifiers mod, bool down) -> Key {
-            return {mod | Key::Modifiers::Scroll,
-                    (Codepoint)((down ? 1 : -1) * m_wheel_scroll_amount)};
+        auto mouse_scroll = [this](Key::Modifiers mod, Codepoint coord, bool down) -> Key {
+            return {mod | Key::Modifiers::Scroll | (Key::Modifiers)((down ? m_wheel_scroll_amount : -1 * m_wheel_scroll_amount) << 16), coord};
         };
 
         auto masked_key = [&](Codepoint key, Codepoint shifted_key = 0) {
@@ -858,6 +861,8 @@ Optional<Key> TerminalUI::get_next_key()
                 return masked_key(Key::F11 + params[0][0] - 23);
             case 25: case 26:
                 return Key{Key::Modifiers::Shift, Key::F3 + params[0][0] - 25}; // rxvt style
+            case 27:
+                return masked_key(convert(static_cast<Codepoint>(params[2][0])));
             case 28: case 29:
                 return Key{Key::Modifiers::Shift, Key::F5 + params[0][0] - 28}; // rxvt style
             case 31: case 32:
@@ -877,8 +882,42 @@ Optional<Key> TerminalUI::get_next_key()
             }
             return {};
         case 'u':
-            return masked_key(convert(static_cast<Codepoint>(params[0][0])),
-                              convert(static_cast<Codepoint>(params[0][1])));
+        {
+            Codepoint key;
+            switch (params[0][0])
+            {
+                // Treat numpad keys the same as their non-numpad counterparts. Could add a numpad modifier here.
+                case 57399: key = '0'; break;
+                case 57400: key = '1'; break;
+                case 57401: key = '2'; break;
+                case 57402: key = '3'; break;
+                case 57403: key = '4'; break;
+                case 57404: key = '5'; break;
+                case 57405: key = '6'; break;
+                case 57406: key = '7'; break;
+                case 57407: key = '8'; break;
+                case 57408: key = '9'; break;
+                case 57409: key = '.'; break;
+                case 57410: key = '/'; break;
+                case 57411: key = '*'; break;
+                case 57412: key = '-'; break;
+                case 57413: key = '+'; break;
+                case 57414: key = Key::Return; break;
+                case 57415: key = '='; break;
+                case 57417: key = Key::Left; break;
+                case 57418: key = Key::Right; break;
+                case 57419: key = Key::Up; break;
+                case 57420: key = Key::Down; break;
+                case 57421: key = Key::PageUp; break;
+                case 57422: key = Key::PageDown; break;
+                case 57423: key = Key::Home; break;
+                case 57424: key = Key::End; break;
+                case 57425: key = Key::Insert; break;
+                case 57426: key = Key::Delete; break;
+                default: key = convert(static_cast<Codepoint>(params[0][0])); break;
+            }
+            return masked_key(key, convert(static_cast<Codepoint>(params[0][1])));
+        }
         case 'Z': return shift(Key::Tab);
         case 'I': return {Key::FocusIn};
         case 'O': return {Key::FocusOut};
@@ -902,8 +941,8 @@ Optional<Key> TerminalUI::get_next_key()
                 else if (int guess = ffs(m_mouse_state) - 1; 0 <= guess and guess < 3)
                     return mouse_button(mod, Key::MouseButton{guess}, coord, true);
                 break;
-            case 64: return mouse_scroll(mod, false);
-            case 65: return mouse_scroll(mod, true);
+            case 64: return mouse_scroll(mod, coord, false);
+            case 65: return mouse_scroll(mod, coord, true);
             }
             return Key{Key::Modifiers::MousePos, coord};
         }
@@ -1292,7 +1331,7 @@ void TerminalUI::info_show(const DisplayLine& title, const DisplayLineList& cont
         max_size.line -= m_menu.size.line;
 
     const auto max_content_width = (m_info_max_width > 0 ? std::min(max_size.column, m_info_max_width) : max_size.column) -
-                                   (framed ? 4 : 2) -
+                                   (framed ? 4 : 0) -
                                    (assisted ? m_assistant[0].column_length() : 0);
     if (max_content_width <= 0)
         return;
@@ -1453,7 +1492,6 @@ void TerminalUI::setup_terminal()
         "\033[22t"    // save the current window title
         "\033[?25l"   // hide cursor
         "\033="       // set application keypad mode, so the keypad keys send unique codes
-        "\033[?2026$p" // query support for synchronize output
         "\033[?2004h" // force enable bracketed-paste events
     );
 }
@@ -1519,10 +1557,15 @@ void TerminalUI::set_ui_options(const Options& options)
 
     m_status_on_top = find("terminal_status_on_top").map(to_bool).value_or(false);
     m_set_title = find("terminal_set_title").map(to_bool).value_or(true);
+    m_title = find("terminal_title").map([](StringView s) { return String{s}; });
 
     auto synchronized = find("terminal_synchronized").map(to_bool);
     m_synchronized.set = (bool)synchronized;
     m_synchronized.requested = synchronized.value_or(false);
+    if (not m_synchronized.queried and not m_synchronized.set) {
+        write(STDOUT_FILENO, "\033[?2026$p");
+        m_synchronized.queried = true;
+    }
 
     m_shift_function_key = find("terminal_shift_function_key").map(str_to_int_ifp).value_or(default_shift_function_key);
 
